/* Memory bounds …
 * © 2010 Heiko Schlittermann <hs@schlittermann.de>
 * based on a idea (see the first revision if you've access to the
 * HG repository) of Frank Epperlein.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <signal.h>
#include <unistd.h>
#include <setjmp.h>

extern char etext, edata, end;

void strings(const char*, const char*);
void segv_handler(int, siginfo_t*, void*);
void set_sighandler(int, void (*)(int, siginfo_t*, void*));

static sigjmp_buf env;

int main(int argc, char **argv)
{
    char* data = malloc(1024);
    strcpy(data, "HALLO HIER IST DER HEAP");

    printf("First address beyond:\n");
    printf("    program text segment(etext)      %10p\n", &etext);
    printf("    initialized data segment(edata)  %10p\n", &edata);
    printf("    uninitialized data segment (end) %10p\n", &end);
    printf("    size: %d\n", &end - &etext);
    printf("Heap near                            %10p\n", data);
    printf("Stack near                           %10p\n", &argc);
    printf("                                   … %10p\n", &data);

    if (isatty(1)) {
	printf("- <ENTER> -------------------------------------\n\n");
	getchar();
    }
    else printf("------------------------------------------------\n\n");

    set_sighandler(SIGSEGV, segv_handler);

    puts("etext -> end");
    strings(&etext, &end);

    puts("end -> …");
    strings(&end, NULL);

    puts("heap -> …");
    strings(data, NULL);

    puts("stack -> …");
    strings((char*) &argc, NULL);

    puts("done");
    return EXIT_SUCCESS;
}

/*
 * No user serviceable parts below this line ☺
 */

void strings(const char *a, const char *b)
{
    const char *p = a;

    /* we try to catch the segv signal using setjmp/longjmp,
     * the setjmp() returns 0 for first invocation, but *seems* to
     * return a 2nd time if somebody called longjmp (as we do inside the
     * signal handler). longjmp() passes the value that magically
     * appears as a result of setjmp().
     * It's really magic ☺
     */

    if (sigsetjmp(env, 1) == 0) {  
	while (b ? p < b : 1) {
	    if (isprint(*p)) {
		printf("%p: ", p);
		while ((b ? p < b : 1) && isprint(*p)) putchar(*p++);
		putchar('\n');
	    }

	    const char *p0 = p;
	    printf("%p: … ", p);
	    while ((b ? p < b : 1) && !isprint(*p)) p++;
	    printf("+%d\n", p - p0);
	}
    }

    printf("%p: <last usable address>\n", p - 1);
}

void segv_handler(int signal, siginfo_t *si, void *p)
{
    printf("segv at address %p\n", si->si_addr);
    siglongjmp(env, 1);
}

void set_sighandler(int signal, void (handler)(int, siginfo_t*, void*))
{
    struct sigaction sa = {  };	    // weiss gerade nicht, ob das eine
				    // gueltige initialisierung ist
    // memset(&sa, 0, sizeof(sa));  // <- das geht in jedem Falle
 
    sa.sa_flags = SA_SIGINFO;
    sa.sa_sigaction = handler;
    if (sigaction(signal, &sa, NULL) < 0) {
	perror("sigaction");
	exit(EXIT_FAILURE);
    }
}

